import pytest


@pytest.mark.sphinx(
    "html",
    testroot="docstrings",
    confoverrides={
        "extensions": ["conflicting_sphinx_extension", "sphinxcontrib_django"]
    },
)
def test_setup_with_conflicting_extension(app, caplog):
    setup_records = caplog.get_records("setup")
    assert len(setup_records) == 1
    assert setup_records[0].name == "sphinxcontrib_django.roles"
    assert "Unable to register cross-reference type: " in setup_records[0].msg



import pytest


@pytest.mark.sphinx("html", testroot="docstrings")
def test_model_method_display(app, do_autodoc):
    actual = do_autodoc(
        app, "method", "dummy_django_app.models.SimpleModel.get_dummy_field_display"
    )
    print(actual)
    assert list(actual) == [
        "",
        ".. py:method:: SimpleModel.get_dummy_field_display()",
        "   :module: dummy_django_app.models",
        "",
        (
            "   Shows the label of the :attr:`dummy_field`. See"
            " :meth:`~django.db.models.Model.get_FOO_display` for more information."
        ),
        "",
    ]


@pytest.mark.sphinx("html", testroot="docstrings")
def test_model_method_get_next_by(app, do_autodoc):
    actual = do_autodoc(
        app, "method", "dummy_django_app.models.SimpleModel.get_next_by_dummy_field"
    )
    print(actual)
    assert list(actual) == [
        "",
        ".. py:method:: SimpleModel.get_next_by_dummy_field()",
        "   :module: dummy_django_app.models",
        "",
        (
            "   Finds next instance based on :attr:`dummy_field`. See"
            " :meth:`~django.db.models.Model.get_next_by_FOO` for more information."
        ),
        "",
    ]


@pytest.mark.sphinx("html", testroot="docstrings")
def test_model_method_get_previous_by(app, do_autodoc):
    actual = do_autodoc(
        app, "method", "dummy_django_app.models.SimpleModel.get_previous_by_dummy_field"
    )
    print(actual)
    assert list(actual) == [
        "",
        ".. py:method:: SimpleModel.get_previous_by_dummy_field()",
        "   :module: dummy_django_app.models",
        "",
        (
            "   Finds previous instance based on :attr:`dummy_field`. See"
            " :meth:`~django.db.models.Model.get_previous_by_FOO` for more information."
        ),
        "",
    ]



import pytest


@pytest.mark.sphinx("html", testroot="docstrings")
def test_data(app, do_autodoc):
    options = {"members": None}
    actual = do_autodoc(app, "module", "dummy_django_app.settings", options)
    print(actual)
    assert list(actual) == [
        "",
        ".. py:module:: dummy_django_app.settings",
        "",
        "   Dummy Django settings file",
        "",
        "",
        ".. py:data:: INSTALLED_APPS",
        "   :module: dummy_django_app.settings",
        (
            "   :value: ['django.contrib.auth', 'django.contrib.contenttypes',"
            " 'dummy_django_app']"
        ),
        "",
        "   These are the installed apps",
        "",
        "   .. code-block:: JavaScript",
        "",
        (
            "       ['django.contrib.auth', 'django.contrib.contenttypes',"
            " 'dummy_django_app']\n"
        ),
        "",
    ]



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

from sphinxcontrib_django import __version__

# -- Project information -----------------------------------------------------

project = "sphinxcontrib-django"
copyright = "2021"
author = "Timo Ludwig"

# The full version, including alpha/beta/rc tags
release = __version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx_rtd_theme",
    "sphinx_last_updated_by_git",
    "sphinxcontrib_django.roles",
]

# Warn about all references where the target cannot be found
nitpicky = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build"]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.
html_theme = "sphinx_rtd_theme"
# The logos shown in the menu bar
html_logo = "images/django-sphinx-logo-white.png"
# The favicon of the html doc files
html_favicon = "images/favicon.svg"
# Do not include links to the documentation source (.rst files) in build
html_show_sourcelink = False



"""
The recommended setup for testing sphinx extension is undocumented (see issue #7008:
https://github.com/sphinx-doc/sphinx/issues/7008), so this setup was created using the given code
snippets and the existing test cases for the autodoc extension.
"""
from unittest.mock import Mock

import pytest
from sphinx.ext.autodoc.directive import DocumenterBridge, process_documenter_options
from sphinx.testing.path import path
from sphinx.util.docutils import LoggingReporter

pytest_plugins = "sphinx.testing.fixtures"


@pytest.fixture(scope="session")
def rootdir():
    """
    Path to the root directory of the testing targets
    """
    return path(__file__).parent.abspath() / "roots"


@pytest.fixture(scope="function")
def app(test_params, app_params, make_app, shared_result, requests_mock):
    """
    Overwrite sphinx.testing.fixtures.app to take the additional fixture requests_mock to fake
    intersphinx requests
    """
    args, kwargs = app_params
    app_ = make_app(*args, **kwargs)
    yield app_


@pytest.fixture(scope="function")
def setup_app_with_different_config(app_params, make_app):
    """
    Simulate the setup of the sphinx app with a different config
    Return the function instead of the final app object to make sure exceptions occur inside the
    test and not inside the fixture
    """

    def setup_app_with_different_config(**confoverrides):
        args, kwargs = app_params
        kwargs["confoverrides"] = confoverrides
        return make_app(*args, **kwargs)

    return setup_app_with_different_config


@pytest.fixture(scope="function")
def do_autodoc():
    """
    This function simulates the autodoc functionality.

    Taken from https://github.com/sphinx-doc/sphinx/blob/d635d94eebbca0ebb1a5402aa07ed58c0464c6d3/tests/test_ext_autodoc.py#L33-L45 # noqa: E501
    """

    def do_autodoc(app, objtype, name, options=None):
        if options is None:
            options = {}
        app.env.temp_data.setdefault("docname", "index")  # set dummy docname
        doccls = app.registry.documenters[objtype]
        docoptions = process_documenter_options(doccls, app.config, options)
        state = Mock()
        state.document.settings.tab_width = 8
        bridge = DocumenterBridge(app.env, LoggingReporter(""), docoptions, 1, state)
        documenter = doccls(bridge, name)
        documenter.generate()
        return bridge.result

    return do_autodoc



import pytest


@pytest.mark.sphinx("html", testroot="docstrings")
def test_exclude_members(app, do_autodoc):
    options = {"members": None, "undoc-members": None}
    assert not any(
        "base_fields" in line
        for line in do_autodoc(app, "module", "dummy_django_app.forms", options)
    )


@pytest.mark.sphinx("html", testroot="docstrings")
def test_include_members(app, do_autodoc):
    options = {"members": None}
    assert any(
        "__init__" in line
        for line in do_autodoc(app, "module", "dummy_django_app.forms", options)
    )



from __future__ import annotations

from django import forms

from .models import SimpleModel


class SimpleForm(forms.ModelForm):
    test1 = forms.CharField(label="Test1")
    test2 = forms.CharField(help_text="Test2")

    def __init__(self, *args, **kwargs) -> None:
        """
        This is a custom init method
        """
        super().__init__(*args, **kwargs)

    class Meta:
        model = SimpleModel
        fields = ("file", "childA", "childrenB", "dummy_field")



"""
Dummy Django settings file
"""

SECRET_KEY = "dummy-key"

#: These are the installed apps
INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "dummy_django_app",
]

USE_TZ = False



import pytest


@pytest.mark.sphinx("html", testroot="docstrings")
def test_simple_model(app, do_autodoc):
    actual = do_autodoc(app, "class", "dummy_django_app.models.SimpleModel")
    print(actual)
    assert list(actual) == [
        "",
        ".. py:class:: SimpleModel(id, file, childA, dummy_field)",
        "   :module: dummy_django_app.models",
        "",
        "   :param id: Primary key: ID",
        "   :type id: ~django.db.models.AutoField",
        "   :param dummy_field: Very verbose name of dummy field. This should help you",
        "",
        "                       Docstring of char field",
        "",
        "                       .. warning::",
        "",
        "                           Inline directives should be preserved.",
        "",
        "   :type dummy_field: ~django.db.models.CharField",
        "",
        "   Relationship fields:",
        "",
        (
            "   :param file: File (related name:"
            " :attr:`~dummy_django_app.models.FileModel.simple_models`)"
        ),
        "",
        "                Docstring of foreign key",
        "",
        (
            "   :type file: :class:`~django.db.models.ForeignKey` to"
            " :class:`~dummy_django_app.models.FileModel`"
        ),
        (
            "   :param childA: ChildA (related name:"
            " :attr:`~dummy_django_app.models.ChildModelA.simple_model`)"
        ),
        "",
        "                  Docstring of one to one field",
        "",
        (
            "   :type childA: :class:`~django.db.models.OneToOneField` to"
            " :class:`~dummy_django_app.models.ChildModelA`"
        ),
        (
            "   :param childrenB: ChildrenB (related name:"
            " :attr:`~dummy_django_app.models.ChildModelB.simple_models`)"
        ),
        "",
        "                     Docstring of many to many field",
        "",
        "                     .. note::",
        "",
        "                         This syntax is also supported.",
        "",
        (
            "   :type childrenB: :class:`~django.db.models.ManyToManyField` to"
            " :class:`~dummy_django_app.models.ChildModelB`"
        ),
        "",
        "   Reverse relationships:",
        "",
        (
            "   :param childmodela: All child model as of this simple model (related"
            " name of :attr:`~dummy_django_app.models.ChildModelA.simple_model`)"
        ),
        (
            "   :type childmodela: Reverse :class:`~django.db.models.ForeignKey` from"
            " :class:`~dummy_django_app.models.ChildModelA`"
        ),
        (
            "   :param childmodelb: All child model bs of this simple model (related"
            " name of :attr:`~dummy_django_app.models.ChildModelB.simple_model`)"
        ),
        (
            "   :type childmodelb: Reverse :class:`~django.db.models.ForeignKey` from"
            " :class:`~dummy_django_app.models.ChildModelB`"
        ),
        "",
        "   .. inheritance-diagram:: dummy_django_app.models.SimpleModel",
        "",
    ]


@pytest.mark.sphinx(
    "html", testroot="docstrings", confoverrides={"django_show_db_tables": True}
)
def test_database_table(app, do_autodoc):
    actual = do_autodoc(app, "class", "dummy_django_app.models.SimpleModel")
    print(actual)
    assert list(actual) == [
        "",
        ".. py:class:: SimpleModel(id, file, childA, dummy_field)",
        "   :module: dummy_django_app.models",
        "",
        "   **Database table:** ``dummy_django_app_simplemodel``",
        "",
        "   :param id: Primary key: ID",
        "   :type id: ~django.db.models.AutoField",
        "   :param dummy_field: Very verbose name of dummy field. This should help you",
        "",
        "                       Docstring of char field",
        "",
        "                       .. warning::",
        "",
        "                           Inline directives should be preserved.",
        "",
        "   :type dummy_field: ~django.db.models.CharField",
        "",
        "   Relationship fields:",
        "",
        (
            "   :param file: File (related name:"
            " :attr:`~dummy_django_app.models.FileModel.simple_models`)"
        ),
        "",
        "                Docstring of foreign key",
        "",
        (
            "   :type file: :class:`~django.db.models.ForeignKey` to"
            " :class:`~dummy_django_app.models.FileModel`"
        ),
        (
            "   :param childA: ChildA (related name:"
            " :attr:`~dummy_django_app.models.ChildModelA.simple_model`)"
        ),
        "",
        "                  Docstring of one to one field",
        "",
        (
            "   :type childA: :class:`~django.db.models.OneToOneField` to"
            " :class:`~dummy_django_app.models.ChildModelA`"
        ),
        (
            "   :param childrenB: ChildrenB (related name:"
            " :attr:`~dummy_django_app.models.ChildModelB.simple_models`)"
        ),
        "",
        "                     Docstring of many to many field",
        "",
        "                     .. note::",
        "",
        "                         This syntax is also supported.",
        "",
        (
            "   :type childrenB: :class:`~django.db.models.ManyToManyField` to"
            " :class:`~dummy_django_app.models.ChildModelB`"
        ),
        "",
        "   Reverse relationships:",
        "",
        (
            "   :param childmodela: All child model as of this simple model (related"
            " name of :attr:`~dummy_django_app.models.ChildModelA.simple_model`)"
        ),
        (
            "   :type childmodela: Reverse :class:`~django.db.models.ForeignKey` from"
            " :class:`~dummy_django_app.models.ChildModelA`"
        ),
        (
            "   :param childmodelb: All child model bs of this simple model (related"
            " name of :attr:`~dummy_django_app.models.ChildModelB.simple_model`)"
        ),
        (
            "   :type childmodelb: Reverse :class:`~django.db.models.ForeignKey` from"
            " :class:`~dummy_django_app.models.ChildModelB`"
        ),
        "",
        "   .. inheritance-diagram:: dummy_django_app.models.SimpleModel",
        "",
    ]


@pytest.mark.sphinx("html", testroot="docstrings")
def test_abstract_model(app, do_autodoc):
    actual = do_autodoc(app, "class", "dummy_django_app.models.AbstractModel")
    print(actual)
    assert list(actual) == [
        "",
        ".. py:class:: AbstractModel(*args, **kwargs)",
        "   :module: dummy_django_app.models",
        "",
        "",
        "   Relationship fields:",
        "",
        "   :param simple_model: Simple model",
        (
            "   :type simple_model: :class:`~django.db.models.ForeignKey` to"
            " :class:`~dummy_django_app.models.SimpleModel`"
        ),
        "   :param user: User",
        (
            "   :type user: :class:`~django.db.models.ForeignKey` to"
            " :class:`~django.contrib.auth.models.User`"
        ),
        "   :param foreignkey_self: Foreignkey self",
        (
            "   :type foreignkey_self: :class:`~django.db.models.ForeignKey` to"
            " :class:`~dummy_django_app.models.AbstractModel`"
        ),
        "",
        "   .. inheritance-diagram:: dummy_django_app.models.AbstractModel",
        "",
    ]


@pytest.mark.sphinx(
    "html", testroot="docstrings", confoverrides={"django_show_db_tables": True}
)
def test_abstract_model_with_tables_names_and_ignore_abstract(app, do_autodoc):
    actual = do_autodoc(app, "class", "dummy_django_app.models.AbstractModel")
    print(actual)
    assert list(actual) == [
        "",
        ".. py:class:: AbstractModel(*args, **kwargs)",
        "   :module: dummy_django_app.models",
        "",
        "",
        "   Relationship fields:",
        "",
        "   :param simple_model: Simple model",
        (
            "   :type simple_model: :class:`~django.db.models.ForeignKey` to"
            " :class:`~dummy_django_app.models.SimpleModel`"
        ),
        "   :param user: User",
        (
            "   :type user: :class:`~django.db.models.ForeignKey` to"
            " :class:`~django.contrib.auth.models.User`"
        ),
        "   :param foreignkey_self: Foreignkey self",
        (
            "   :type foreignkey_self: :class:`~django.db.models.ForeignKey` to"
            " :class:`~dummy_django_app.models.AbstractModel`"
        ),
        "",
        "   .. inheritance-diagram:: dummy_django_app.models.AbstractModel",
        "",
    ]


@pytest.mark.sphinx(
    "html",
    testroot="docstrings",
    confoverrides={
        "django_show_db_tables": True,
        "django_show_db_tables_abstract": True,
    },
)
def test_abstract_model_with_tables_names_and_abstract_show(app, do_autodoc):
    actual = do_autodoc(app, "class", "dummy_django_app.models.AbstractModel")
    print(actual)
    assert list(actual) == [
        "",
        ".. py:class:: AbstractModel(*args, **kwargs)",
        "   :module: dummy_django_app.models",
        "",
        "   **Database table:** ``None``",
        "",
        "",
        "   Relationship fields:",
        "",
        "   :param simple_model: Simple model",
        (
            "   :type simple_model: :class:`~django.db.models.ForeignKey` to"
            " :class:`~dummy_django_app.models.SimpleModel`"
        ),
        "   :param user: User",
        (
            "   :type user: :class:`~django.db.models.ForeignKey` to"
            " :class:`~django.contrib.auth.models.User`"
        ),
        "   :param foreignkey_self: Foreignkey self",
        (
            "   :type foreignkey_self: :class:`~django.db.models.ForeignKey` to"
            " :class:`~dummy_django_app.models.AbstractModel`"
        ),
        "",
        "   .. inheritance-diagram:: dummy_django_app.models.AbstractModel",
        "",
    ]


@pytest.mark.sphinx("html", testroot="docstrings")
def test_file_model(app, do_autodoc):
    actual = do_autodoc(app, "class", "dummy_django_app.models.FileModel")
    print(actual)
    assert list(actual) == [
        "",
        ".. py:class:: FileModel(id, upload)",
        "   :module: dummy_django_app.models",
        "",
        "   :param id: Primary key: ID",
        "   :type id: ~django.db.models.AutoField",
        "   :param upload: Upload",
        "   :type upload: ~django.db.models.FileField",
        "",
        "   Reverse relationships:",
        "",
        (
            "   :param simple_models: All simple models of this file model (related"
            " name of :attr:`~dummy_django_app.models.SimpleModel.file`)"
        ),
        (
            "   :type simple_models: Reverse :class:`~django.db.models.ForeignKey` from"
            " :class:`~dummy_django_app.models.SimpleModel`"
        ),
        "",
        "   .. inheritance-diagram:: dummy_django_app.models.FileModel",
        "",
    ]


@pytest.mark.sphinx("html", testroot="docstrings")
def test_tagged_item(app, do_autodoc):
    actual = do_autodoc(app, "class", "dummy_django_app.models.TaggedItem")
    print(actual)
    assert list(actual) == [
        "",
        ".. py:class:: TaggedItem(id, tag, content_type, object_id)",
        "   :module: dummy_django_app.models",
        "",
        "   :param id: Primary key: ID",
        "   :type id: ~django.db.models.AutoField",
        "   :param tag: Tag",
        "   :type tag: ~django.db.models.SlugField",
        "   :param object_id: Object id",
        "   :type object_id: ~django.db.models.PositiveIntegerField",
        (
            "   :param content_object: Generic foreign key to the"
            " :class:`~django.contrib.contenttypes.models.ContentType` specified in"
            " :attr:`~dummy_django_app.models.TaggedItem.content_type`"
        ),
        (
            "   :type content_object:"
            " ~django.contrib.contenttypes.fields.GenericForeignKey"
        ),
        "",
        "   Relationship fields:",
        "",
        (
            "   :param content_type: Content type (related name:"
            " :attr:`~django.contrib.contenttypes.models.ContentType.taggeditem`)"
        ),
        (
            "   :type content_type: :class:`~django.db.models.ForeignKey` to"
            " :class:`~django.contrib.contenttypes.models.ContentType`"
        ),
        "",
        "   .. inheritance-diagram:: dummy_django_app.models.TaggedItem",
        "",
    ]


@pytest.mark.sphinx("html", testroot="docstrings")
def test_form(app, do_autodoc):
    actual = do_autodoc(app, "class", "dummy_django_app.forms.SimpleForm")
    print(actual)
    assert list(actual) == [
        "",
        ".. py:class:: SimpleForm(*args, **kwargs)",
        "   :module: dummy_django_app.forms",
        "",
        "   **Form fields:**",
        "",
        "   * ``file``: File (:class:`~django.forms.ModelChoiceField`)",
        "   * ``childA``: ChildA (:class:`~django.forms.ModelChoiceField`)",
        (
            "   * ``childrenB``: ChildrenB"
            " (:class:`~django.forms.ModelMultipleChoiceField`)"
        ),
        (
            "   * ``dummy_field``: Very verbose name of dummy field"
            " (:class:`~django.forms.CharField`)"
        ),
        "   * ``test1``: Test1 (:class:`~django.forms.CharField`)",
        "   * ``test2``: Test2 (:class:`~django.forms.CharField`)",
        "",
    ]



import pytest


@pytest.mark.sphinx("html", testroot="docstrings")
def test_django_configured(app, do_autodoc):
    actual = do_autodoc(app, "class", "dummy_django_app.models.MonkeyPatched")
    print(actual)
    assert list(actual) == [
        "",
        ".. py:class:: MonkeyPatched(*args, **kwargs)",
        "   :module: dummy_django_app.models",
        "",
        "   Monkeypatched docstring",
        "",
        "   :param id: Primary key: ID",
        "   :type id: ~django.db.models.AutoField",
        "",
        "   .. inheritance-diagram:: dummy_django_app.models.MonkeyPatched",
        "",
    ]



import pytest

try:
    from phonenumber_field.modelfields import PhoneNumberField  # noqa: F401

    PHONENUMBER = True
except ModuleNotFoundError:
    # In case phonenumber is not used, pass
    PHONENUMBER = False


@pytest.mark.sphinx("html", testroot="docstrings")
def test_model_field(app, do_autodoc):
    actual = do_autodoc(
        app, "attribute", "dummy_django_app.models.SimpleModel.dummy_field"
    )
    print(actual)
    assert list(actual) == [
        "",
        ".. py:attribute:: SimpleModel.dummy_field",
        "   :module: dummy_django_app.models",
        "",
        "   Type: :class:`~django.db.models.CharField`",
        "",
        "   Very verbose name of dummy field. This should help you",
        "",
        "   Docstring of char field",
        "",
        "   .. warning::",
        "",
        "       Inline directives should be preserved.",
        "",
    ]


@pytest.mark.sphinx("html", testroot="docstrings")
def test_foreignkey(app, do_autodoc):
    actual = do_autodoc(app, "attribute", "dummy_django_app.models.SimpleModel.file")
    print(actual)
    assert list(actual) == [
        "",
        ".. py:attribute:: SimpleModel.file",
        "   :module: dummy_django_app.models",
        "",
        (
            "   Type: :class:`~django.db.models.ForeignKey` to"
            " :class:`~dummy_django_app.models.FileModel`"
        ),
        "",
        (
            "   File (related name:"
            " :attr:`~dummy_django_app.models.FileModel.simple_models`)"
        ),
        "",
        "   Docstring of foreign key",
        "",
    ]


@pytest.mark.sphinx("html", testroot="docstrings")
def test_foreignkey_id(app, do_autodoc):
    actual = do_autodoc(app, "attribute", "dummy_django_app.models.SimpleModel.file_id")
    print(actual)
    assert list(actual) == [
        "",
        ".. py:attribute:: SimpleModel.file_id",
        "   :module: dummy_django_app.models",
        "",
        (
            "   Internal field, use :class:`~dummy_django_app.models.SimpleModel.file`"
            " instead."
        ),
        "",
    ]


@pytest.mark.sphinx("html", testroot="docstrings")
def test_foreignkey_string(app, do_autodoc):
    actual = do_autodoc(app, "attribute", "dummy_django_app.models.SimpleModel.file")
    print(actual)
    assert list(actual) == [
        "",
        ".. py:attribute:: SimpleModel.file",
        "   :module: dummy_django_app.models",
        "",
        (
            "   Type: :class:`~django.db.models.ForeignKey` to"
            " :class:`~dummy_django_app.models.FileModel`"
        ),
        "",
        (
            "   File (related name:"
            " :attr:`~dummy_django_app.models.FileModel.simple_models`)"
        ),
        "",
        "   Docstring of foreign key",
        "",
    ]


@pytest.mark.sphinx("html", testroot="docstrings")
def test_foreignkey_string_abstract_model(app, do_autodoc):
    actual = do_autodoc(
        app, "attribute", "dummy_django_app.models.AbstractModel.simple_model"
    )
    print(actual)
    assert list(actual) == [
        "",
        ".. py:attribute:: AbstractModel.simple_model",
        "   :module: dummy_django_app.models",
        "",
        (
            "   Type: :class:`~django.db.models.ForeignKey` to"
            " :class:`~dummy_django_app.models.SimpleModel`"
        ),
        "",
        "   Simple model",
        "",
    ]


@pytest.mark.sphinx("html", testroot="docstrings")
def test_reverse_foreignkey(app, do_autodoc):
    actual = do_autodoc(
        app, "attribute", "dummy_django_app.models.FileModel.simple_models"
    )
    print(actual)
    assert list(actual) == [
        "",
        ".. py:attribute:: FileModel.simple_models",
        "   :module: dummy_django_app.models",
        "",
        (
            "   Type: Reverse :class:`~django.db.models.ForeignKey` from"
            " :class:`~dummy_django_app.models.SimpleModel`"
        ),
        "",
        (
            "   All simple models of this file model (related name of"
            " :attr:`~dummy_django_app.models.SimpleModel.file`)"
        ),
        "",
    ]


@pytest.mark.sphinx("html", testroot="docstrings")
def test_manytomany_field(app, do_autodoc):
    actual = do_autodoc(
        app, "attribute", "dummy_django_app.models.SimpleModel.childrenB"
    )
    print(actual)
    assert list(actual) == [
        "",
        ".. py:attribute:: SimpleModel.childrenB",
        "   :module: dummy_django_app.models",
        "",
        (
            "   Type: :class:`~django.db.models.ManyToManyField` to"
            " :class:`~dummy_django_app.models.ChildModelB`"
        ),
        "",
        (
            "   ChildrenB (related name:"
            " :attr:`~dummy_django_app.models.ChildModelB.simple_models`)"
        ),
        "",
        "   Docstring of many to many field",
        "",
        "   .. note::",
        "",
        "       This syntax is also supported.",
        "",
    ]


@pytest.mark.sphinx("html", testroot="docstrings")
def test_reverse_manytomany_field(app, do_autodoc):
    actual = do_autodoc(
        app, "attribute", "dummy_django_app.models.ChildModelB.simple_models"
    )
    print(actual)
    assert list(actual) == [
        "",
        ".. py:attribute:: ChildModelB.simple_models",
        "   :module: dummy_django_app.models",
        "",
        (
            "   Type: Reverse :class:`~django.db.models.ManyToManyField` from"
            " :class:`~dummy_django_app.models.SimpleModel`"
        ),
        "",
        (
            "   All simple models of this child model b (related name of"
            " :attr:`~dummy_django_app.models.SimpleModel.childrenB`)"
        ),
        "",
    ]


@pytest.mark.sphinx("html", testroot="docstrings")
def test_reverse_onetoone_field(app, do_autodoc):
    actual = do_autodoc(
        app, "attribute", "dummy_django_app.models.ChildModelA.simple_model"
    )
    print(actual)
    assert list(actual) == [
        "",
        ".. py:attribute:: ChildModelA.simple_model",
        "   :module: dummy_django_app.models",
        "",
        (
            "   Type: Reverse :class:`~django.db.models.OneToOneField` from"
            " :class:`~dummy_django_app.models.SimpleModel`"
        ),
        "",
        (
            "   The simple model of this child model a (related name of"
            " :attr:`~dummy_django_app.models.SimpleModel.childA`)"
        ),
        "",
    ]


@pytest.mark.sphinx("html", testroot="docstrings")
def test_generic_foreign_key(app, do_autodoc):
    actual = do_autodoc(
        app, "attribute", "dummy_django_app.models.TaggedItem.content_object"
    )
    print(actual)
    assert list(actual) == [
        "",
        ".. py:attribute:: TaggedItem.content_object",
        "   :module: dummy_django_app.models",
        "",
        "   Provide a generic many-to-one relation through the ``content_type`` and",
        "   ``object_id`` fields.",
        "",
        "   This class also doubles as an accessor to the related object (similar to",
        "   ForwardManyToOneDescriptor) by adding itself as a model attribute.",
        "",
    ]


@pytest.mark.sphinx("html", testroot="docstrings")
def test_model_manager_fields(app, do_autodoc):
    actual = do_autodoc(
        app, "attribute", "dummy_django_app.models.SimpleModel.custom_objects"
    )
    print(actual)
    assert list(actual) == [
        "",
        ".. py:attribute:: SimpleModel.custom_objects",
        "   :module: dummy_django_app.models",
        "   :value: <dummy_django_app.models.SimpleModelManager object>",
        "",
        "   Django manager to access the ORM",
        "   Use ``SimpleModel.objects.all()`` to fetch all objects.",
        "",
        "   Custom model manager",
        "",
    ]


@pytest.mark.sphinx("html", testroot="docstrings")
def test_file_field(app, do_autodoc):
    actual = do_autodoc(app, "attribute", "dummy_django_app.models.FileModel.upload")
    print(actual)
    assert list(actual) == [
        "",
        ".. py:attribute:: FileModel.upload",
        "   :module: dummy_django_app.models",
        "",
        "   Type: :class:`~django.db.models.FileField`",
        "",
        "   Upload",
        "",
    ]


@pytest.mark.sphinx("html", testroot="docstrings")
def test_choice_field(app, do_autodoc):
    actual = do_autodoc(
        app, "attribute", "dummy_django_app.models.ChoiceModel.choice_limit_below"
    )
    print(actual)
    assert list(actual) == [
        "",
        ".. py:attribute:: ChoiceModel.choice_limit_below",
        "   :module: dummy_django_app.models",
        "",
        "   Type: :class:`~django.db.models.IntegerField`",
        "",
        "   Choice limit below",
        "",
        "   Choices:",
        "",
        "   * ``0``",
        "   * ``1``",
        "   * ``2``",
        "   * ``3``",
        "   * ``4``",
        "   * ``5``",
        "   * ``6``",
        "   * ``7``",
        "   * ``8``",
        "",
    ]


@pytest.mark.sphinx("html", testroot="docstrings")
def test_choice_field_limit_exact(app, do_autodoc):
    actual = do_autodoc(
        app, "attribute", "dummy_django_app.models.ChoiceModel.choice_limit_exact"
    )
    print(actual)
    assert list(actual) == [
        "",
        ".. py:attribute:: ChoiceModel.choice_limit_exact",
        "   :module: dummy_django_app.models",
        "",
        "   Type: :class:`~django.db.models.IntegerField`",
        "",
        "   Choice limit exact",
        "",
        "   Choices:",
        "",
        "   * ``0``",
        "   * ``1``",
        "   * ``2``",
        "   * ``3``",
        "   * ``4``",
        "   * ``5``",
        "   * ``6``",
        "   * ``7``",
        "   * ``8``",
        "   * ``9``",
        "   * ``10``",
        "",
    ]


@pytest.mark.sphinx("html", testroot="docstrings")
def test_choice_field_limit_above(app, do_autodoc):
    actual = do_autodoc(
        app, "attribute", "dummy_django_app.models.ChoiceModel.choice_limit_above"
    )
    print(actual)
    assert list(actual) == [
        "",
        ".. py:attribute:: ChoiceModel.choice_limit_above",
        "   :module: dummy_django_app.models",
        "",
        "   Type: :class:`~django.db.models.IntegerField`",
        "",
        "   Choice limit above",
        "",
        "   Choices:",
        "",
        "   * ``0``",
        "   * ``1``",
        "   * ``2``",
        "   * ``3``",
        "   * ``4``",
        "   * ``5``",
        "   * ``6``",
        "   * ``7``",
        "   * ``8``",
        "   * ``9``",
        "   * and 2 more",
        "",
    ]


@pytest.mark.sphinx(
    "html", testroot="docstrings", confoverrides={"django_choices_to_show": 15}
)
def test_choice_field_custom_limit(app, do_autodoc):
    actual = do_autodoc(
        app, "attribute", "dummy_django_app.models.ChoiceModel.choice_limit_above"
    )
    print(actual)
    assert list(actual) == [
        "",
        ".. py:attribute:: ChoiceModel.choice_limit_above",
        "   :module: dummy_django_app.models",
        "",
        "   Type: :class:`~django.db.models.IntegerField`",
        "",
        "   Choice limit above",
        "",
        "   Choices:",
        "",
        "   * ``0``",
        "   * ``1``",
        "   * ``2``",
        "   * ``3``",
        "   * ``4``",
        "   * ``5``",
        "   * ``6``",
        "   * ``7``",
        "   * ``8``",
        "   * ``9``",
        "   * ``10``",
        "   * ``11``",
        "",
    ]


@pytest.mark.sphinx("html", testroot="docstrings")
def test_choice_field_empty(app, do_autodoc):
    actual = do_autodoc(
        app, "attribute", "dummy_django_app.models.ChoiceModel.choice_with_empty"
    )
    print(actual)
    assert list(actual) == [
        "",
        ".. py:attribute:: ChoiceModel.choice_with_empty",
        "   :module: dummy_django_app.models",
        "",
        "   Type: :class:`~django.db.models.CharField`",
        "",
        "   Choice with empty",
        "",
        "   Choices:",
        "",
        "   * ``''`` (Empty string)",
        "   * ``Something``",
        "",
    ]


if PHONENUMBER:

    @pytest.mark.sphinx("html", testroot="docstrings")
    def test_phonenumber_field(app, do_autodoc):
        actual = do_autodoc(
            app, "attribute", "dummy_django_app.models.PhoneNumberModel.phone_number"
        )
        print(actual)
        assert list(actual) == [
            "",
            ".. py:attribute:: PhoneNumberModel.phone_number",
            "   :module: dummy_django_app.models",
            "",
            "   Type: :class:`~phonenumber_field.modelfields.PhoneNumberField`",
            "",
            "   Phone number",
            "",
        ]



from __future__ import annotations

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models

from sphinxcontrib_django.docstrings.config import CHOICES_LIMIT

try:
    from phonenumber_field.modelfields import PhoneNumberField

    PHONENUMBER = True
except ModuleNotFoundError:
    # In case phonenumber is not used, pass
    PHONENUMBER = False


class SimpleModelManager(models.Manager):
    pass


class FileModel(models.Model):
    upload = models.FileField()


class SimpleModel(models.Model):
    #: Docstring of foreign key
    file = models.ForeignKey(
        "FileModel", related_name="simple_models", on_delete=models.CASCADE
    )

    #: Docstring of one to one field
    childA = models.OneToOneField(
        "ChildModelA", related_name="simple_model", on_delete=models.CASCADE
    )

    childrenB = models.ManyToManyField("ChildModelB", related_name="simple_models")
    """
    Docstring of many to many field

    .. note::

        This syntax is also supported.
    """

    #: Docstring of char field
    #:
    #: .. warning::
    #:
    #:     Inline directives should be preserved.
    dummy_field = models.CharField(
        max_length=3,
        help_text="This should help you",
        verbose_name="Very verbose name of dummy field",
    )

    #: Custom model manager
    custom_objects = SimpleModelManager()

    # Mock get_..._display method of Django models
    def get_dummy_field_display(self):
        pass

    # Mock common get_next_by_ method
    def get_next_by_dummy_field(self):
        pass

    # Mock common get_previous_by_ method
    def get_previous_by_dummy_field(self):
        pass


class AbstractModel(models.Model):
    simple_model = models.ForeignKey("SimpleModel", on_delete=models.CASCADE)
    user = models.ForeignKey("auth.User", related_name="+", on_delete=models.CASCADE)
    foreignkey_self = models.ForeignKey("self", on_delete=models.CASCADE)

    class Meta:
        abstract = True


class ChildModelA(AbstractModel):
    pass


class ChildModelB(AbstractModel):
    pass


class ChoiceModel(models.Model):
    choice_limit_below = models.IntegerField(
        choices=[(i, i) for i in range(CHOICES_LIMIT - 1)]
    )
    choice_limit_exact = models.IntegerField(
        choices=[(i, i) for i in range(CHOICES_LIMIT + 1)]
    )
    choice_limit_above = models.IntegerField(
        choices=[(i, i) for i in range(CHOICES_LIMIT + 2)]
    )
    choice_with_empty = models.CharField(
        choices=[("", "Empty"), ("Something", "Not empty")]
    )


class TaggedItem(models.Model):
    # Test model taken from:
    # https://docs.djangoproject.com/en/stable/ref/contrib/contenttypes/#generic-relations
    tag = models.SlugField()
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey("content_type", "object_id")

    def __str__(self) -> str:
        return self.tag


if PHONENUMBER:

    class PhoneNumberModel(models.Model):
        phone_number = PhoneNumberField()


class MonkeyPatched(models.Model):
    pass



"""
This module contains all functions which are used to improve the documentation of classes.
"""
from __future__ import annotations

from django import forms
from django.db import models
from sphinx.application import Sphinx
from sphinx.pycode import ModuleAnalyzer

from .field_utils import get_field_type, get_field_verbose_name


def improve_class_docstring(app: Sphinx, cls: type, lines: list[str]) -> None:
    """
    Improve the documentation of a class if it's a Django model or form

    :param app: The Sphinx application object
    :param cls: The instance of the class to document
    :param lines: The docstring lines
    """
    if issubclass(cls, models.Model):
        improve_model_docstring(app, cls, lines)
    elif issubclass(cls, forms.BaseForm):
        improve_form_docstring(cls, lines)


def improve_model_docstring(app: Sphinx, model: models.Model, lines: list[str]) -> None:
    """
    Improve the documentation of a Django :class:`~django.db.models.Model` subclass.

    This adds all model fields as parameters to the ``__init__()`` method.

    :param app: The Sphinx application object
    :param model: The instance of the model to document
    :param lines: The docstring lines
    """

    # Add database table name
    if app.config.django_show_db_tables:
        add_db_table_name(app, model, lines)

    # Get predefined params to exclude them from the automatically inserted params
    param_offset = len(":param ")
    predefined_params = [
        line[param_offset : line.find(":", param_offset)]
        for line in lines
        if line.startswith(":param ") and ":" in line[param_offset:]
    ]

    # Get all fields of this model which are not already explicitly included in the docstring
    all_fields = [
        field
        for field in model._meta.get_fields(include_parents=True)
        if field.name not in predefined_params
    ]
    # Get all related fields (ForeignKey, OneToOneField, ManyToManyField)
    related_fields = [
        field
        for field in all_fields
        if isinstance(field, models.fields.related.RelatedField)
    ]
    # Get all reverse relationships
    reverse_related_fields = [
        field
        for field in all_fields
        if isinstance(field, models.fields.reverse_related.ForeignObjectRel)
    ]
    # All fields which are neither related nor reverse related
    non_related_fields = [
        field
        for field in all_fields
        if field not in related_fields + reverse_related_fields
    ]

    # Analyze model to get inline field docstrings
    analyzer = ModuleAnalyzer.for_module(model.__module__)
    analyzer.analyze()
    field_docs = {
        field_name: field_docstring
        for (_, field_name), field_docstring in analyzer.attr_docs.items()
    }

    # Add the normal fields to the docstring
    add_model_parameters(non_related_fields, lines, field_docs)

    # Add the related fields
    if related_fields:
        lines.append("")
        lines.append("Relationship fields:")
        lines.append("")
        add_model_parameters(related_fields, lines, field_docs)

    # Add the reverse related fields
    if reverse_related_fields:
        lines.append("")
        lines.append("Reverse relationships:")
        lines.append("")
        add_model_parameters(reverse_related_fields, lines, field_docs)

    # Add the inheritance diagram
    if (
        "sphinx.ext.inheritance_diagram" in app.extensions
        and "sphinx.ext.graphviz" in app.extensions
        and not any("inheritance-diagram::" in line for line in lines)
    ):
        lines.append("")
        lines.append(f".. inheritance-diagram:: {model.__module__}.{model.__name__}")
        lines.append("")


def add_db_table_name(app: Sphinx, model: models.Model, lines: list[str]) -> None:
    """
    Format and add table name by extension configuration.

    :param app: The Sphinx application object
    :param model: The instance of the model to document
    :param lines: The docstring lines
    """
    if model._meta.abstract and not app.config.django_show_db_tables_abstract:
        return

    table_name = None if model._meta.abstract else model._meta.db_table
    lines.insert(0, "")
    lines.insert(0, f"**Database table:** ``{table_name}``")


def add_model_parameters(
    fields: list[models.Field], lines: list[str], field_docs: dict
) -> None:
    """
    Add the given fields as model parameter with the ``:param:`` directive

    :param fields: The list of fields
    :param lines: The list of current docstring lines
    :param field_docs: The attribute docstrings of the model
    """
    for field in fields:
        # Add docstrings if they are found
        docstring_lines = field_docs.get(field.name, [])
        # Add param doc line
        param = f":param {field.name}: "
        lines.append(param + get_field_verbose_name(field))
        if docstring_lines:
            # Separate from verbose name
            lines.append("")
        # Add and indent existing docstring lines
        lines.extend([(" " * len(param)) + line for line in docstring_lines])

        # Add type
        lines.append(f":type {field.name}: {get_field_type(field, include_role=False)}")


def improve_form_docstring(form: forms.Form, lines: list[str]) -> None:
    """
    Improve the documentation of a Django :class:`~django.forms.Form` class.
    This highlights the available fields in the form.

    :param form: The form object
    :param lines: The list of existing docstring lines
    """
    lines.append("**Form fields:**")
    lines.append("")
    for name, field in form.base_fields.items():
        field_type = f"{field.__class__.__module__}.{field.__class__.__name__}"
        label = field.label or name.replace("_", " ").title()
        lines.append(f"* ``{name}``: {label} (:class:`~{field_type}`)")



def setup(app):
    """
    Sphinx extension which also registers a "setting" directive
    """
    app.add_crossref_type(
        directivename="setting", rolename="setting", indextemplate="pair: %s; setting"
    )



import os
import sys

# Add directory containing dummy app to sys.path
sys.path.insert(0, os.path.abspath("."))

project = "sphinx dummy Test"
extensions = [
    "sphinxcontrib_django",
    "sphinx.ext.graphviz",
    "sphinx.ext.inheritance_diagram",
]

# Configure Django settings module
django_settings = "dummy_django_app.settings"

nitpicky = True


def patch_django_for_autodoc(app):
    """
    Monkeypatch application
    """
    from dummy_django_app.models import MonkeyPatched

    MonkeyPatched.__doc__ = "Monkeypatched docstring"


def setup(app):
    # Run method after Django config is initialized
    app.connect("django-configured", patch_django_for_autodoc)



import pytest
from sphinx.errors import ConfigError


@pytest.mark.sphinx("html", testroot="docstrings")
def test_setup_with_missing_config(setup_app_with_different_config):
    """
    Simulate the missing configuration of the Django settings
    """
    with pytest.raises(ConfigError):
        setup_app_with_different_config(django_settings="")


@pytest.mark.sphinx("html", testroot="docstrings")
def test_setup_with_incorrect_config(setup_app_with_different_config):
    """
    Simulate the incorrect configuration of the Django settings
    """
    with pytest.raises(ConfigError):
        setup_app_with_different_config(django_settings="non_existing_module.settings")
